# -*- coding: utf-8 -*-
#
#  Cloud-Control Node
#  Copyright (C) 2011 SmartJog [http://www.smartjog.com]
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  Authors:
#    * 2011-06: Thibault VINCENT <thibault.vincent@smartjog.com>
#

from __future__ import absolute_import
import ConfigParser
from optparse import OptionParser
from . import __version__

ARGS_DEFAULTS = {
    'daemonize'         : (bool, False),
    'stdout'            : (bool, False),
    'config'            : (str,  '/etc/cc-node.conf'),
    'pidfile'           : (str,  '/var/run/cc-node.pid'),
}
FILE_DEFAULTS = {
    'address'           : (str,  ''),
    'port'              : (int,  1984),
    'ssl'               : (bool, True),
    'login'             : (str,  ''),
    'password'          : (str,  ''),
    'verbosity'         : (int,  0),
    'virtualization'    : (bool, True),
    'command_execution' : (bool, True),
}
SECURE_OPTIONS = [
    'password',
]


class ConfigurationError(Exception):
    
    pass


class Configuration(object):
    
    def __init__(self, logger):
        
        self._logger = logger
        
        # populate default options
        self._options = {}
        for name, (vtype, value) in FILE_DEFAULTS.iteritems():
            self._options[name] = _ConfigValue(vtype, value)
        for name, (vtype, value) in ARGS_DEFAULTS.iteritems():
            self._options[name] = _ConfigValue(vtype, value, locking=True)
    
    def parse_arguments(self):
        
        parser = OptionParser(version = '%%prog %s' % __version__)
        
        # --daemonize
        parser.add_option('-d',
                          '--daemonize',
                          default = ARGS_DEFAULTS['daemonize'][1],
                          action = 'store_true',
                          help = 'run as daemon and write pid file')
        # --stdout
        parser.add_option('-s',
                          '--stdout',
                          action = 'store_true',
                          default = ARGS_DEFAULTS['stdout'][1],
                          help = 'log on standard output instead of syslog'
                                 ' (when not running in daemon mode)')
        # --config
        parser.add_option('-c',
                          '--config',
                          default = ARGS_DEFAULTS['config'][1],
                          help = 'configuration file path (default: %default)')
        # --pidfile
        parser.add_option('-p',
                          '--pidfile',
                          default = ARGS_DEFAULTS['pidfile'][1],
                          help = 'pid file path for daemon mode (default: '
                                 '%default)')
        
        # parse command line
        self._logger.info('reading command-line arguments')
        options = parser.parse_args()[0]
        
        # apply arguments to current configuration
        for name, value in options.__dict__.iteritems():
            if name in self._options:
                self._options[name].update(value, unlock=True)
                if name not in SECURE_OPTIONS:
                    self._logger.info(' * %s = %s' % (name,
                                            self._options[name].value()))
    
    def parse_file(self):
        
        parser = ConfigParser.SafeConfigParser()
        
        # open configuration file
        try:
            self._logger.info('parsing configuration file')
            parser.read(self._options['config'].value())
            options = dict(parser.items('node'))
        except ConfigParser.Error as err:
            raise ConfigurationError('failed to open or parse configuration'
                                     ' file: %r' % err)
        
        # apply to current configuration
        for name, value in options.iteritems():
            if name in self._options:
                self._options[name].update(value, unlock=False)
                if name not in SECURE_OPTIONS:
                    self._logger.info(' * %s = %s' % (name,
                                            self._options[name].value()))
    
    def get(self, name):
        
        if name in self._options:
            return self._options[name].value()
        else:
            return None
    
    def update(self, name, value):
        
        # TODO missing concurrency protection
        if name in self._options:
            self._options[name].update(value, unlock=False)


class _ConfigValue(object):
    
    def __init__(self, value_type, default_value, locking=False):
        
        self._type = value_type
        self._default = default_value
        self._value = default_value
        self._locking = locking
    
    def __str__(self):
        
        return '' if self._value is None else str(self._value)
    
    def update(self, value, unlock=False):
        
        if not self._locking or (self._locking and unlock):
            try:
                # special case for boolean conversions
                if self._type is bool and not isinstance(value, bool):
                    # any integer != 0 means True
                    if isinstance(value, int):
                        self._value = value != 0
                    # a string meaning 'yes'
                    elif isinstance(value, str):
                        self._value = value.strip().lower() in ('yes',
                                                    'y', '1', 'true', 'enable',
                                                    'enabled', 'active', 'on')
                    # TODO rethink about a default value, or how to warn
                    else:
                        self._value = False
                # or just let python do the conversion
                else:
                    self._value = self._type(value)
            except ValueError:
                raise ConfigurationError('option has type `%r`, cannot'
                            ' convert from value `%s`' % (self._type, value))
            except TypeError:
                raise ConfigurationError('option convertion to `%r` failed'
                                                                % (self._type))
    
    def value(self):
        
        return self._value
    
    def default_value(self):
        
        return self._default
    
    def is_default(self):
        
        return self._value == self._default
