# -*- coding: utf-8 -*-
#
#  Cloud-Control Node
#  Copyright (C) 2011 SmartJog [http://www.smartjog.com]
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  Authors:
#    * 2011-06: Thibault VINCENT <thibault.vincent@smartjog.com>
#

from __future__ import absolute_import
from threading import Lock, Event, Thread
from time import sleep
from sjrpc.client import SimpleRpcClient
from sjrpc.utils import ConnectionProxy, pure
from sjrpc.core import RpcError
from .logging import Logger
from .jobs import JobManager
from .threading import RWLock
from . import modules

pure = pure


class CoreError(Exception):
    
    pass


class Core(object):
    
    def __init__(self, logger, configuration):
        
        self._logger = logger
        self._config = configuration
        self._modmgr = ModuleManager(self._logger, self._config)
        
        # big lock for events synchronization
        self._mutex = Lock()
        # state initialized as shutdown
        self._shutdown = Event()
        self._shutdown.set()

    def run(self):
        
        # never run twice (should not happen, but let's be sure)
        with self._mutex:
            if self._shutdown.is_set():
                self._shutdown.clear()
            else:
                raise CoreError('already running')
        
        # exceptions in this block will force a shutdown
        try:
            ### create rpc client
            rpc_handler = Handler(self._logger, "core")
            rpc_manager = SimpleRpcClient.from_addr(
                                    self._config.get('address'),
                                    self._config.get('port'),
                                    enable_ssl = self._config.get('ssl'),
                                    default_handler = rpc_handler)
            rpc_proxy = ConnectionProxy(rpc_manager)
            # FIXME bad API workaround
            self._rpc_manager = rpc_manager
            
            # FIXME bad API, move that to __init__ and clean handler each time
            #       we re-enter run()
            self._modmgr.set_core_handler(rpc_handler)
            
            ### run the sjrpc main loop in a separate thread
            ### because we need it for authentication
            rpc_thread = Thread(name='sjrpc', target=rpc_manager.run)
            rpc_thread.start()
            
            ### authenticate with the server and set node role
            while rpc_thread.is_alive():
                try:
                    self._logger.debug(1, 'sending authentication request')
                    # send credential and fetch the node role
                    role = rpc_proxy.authentify(self._config.get('login'),
                                                self._config.get('password'))
                # authentication failed
                except RpcError as err:
                    # bad credentials
                    if err.exception == 'AuthenticationError':
                        self._logger.info('authentication failed,'
                                                        ' retrying soon...')
                    # other errors
                    else:
                        self._logger.warning('unknown error during'
                                                    ' authentication: %s', err)
                    # wait before trying again
                    sleep(10)
                                                        
                # authentication succeeded
                else:
                    self._logger.info('authentication with server succeeded')
                    
                    # received hypervisor role: we obey local settings
                    if role == 'hv':
                        self._logger.info('server affected hypervisor role')
                        if not self._config.get('virtualization'):
                            self._logger.warning('virtualization features are'
                                ' locally disabled by configuration, cannot'
                                ' comply with server-affected role')
                                
                    # received host role: disabling virtualization
                    elif role == 'host':
                        self._logger.info('server affected node role,'
                                        ' disabling virtualization features')
                        self._config.update('virtualization', False)
                        
                    # unknown role, core should be restarted to close connection
                    else:
                        self._logger.error('wrong role affected by'
                                                            ' server: %s', role)
                        raise CoreError('wrong role affected')
                    # quit auth loop
                    break
            
            ### check a last time that sjrpc is running 
            if not rpc_thread.is_alive():
                self._logger.critical('sjrpc main loop terminated too soon')
                raise CoreError('lost sjrpc main loop')
            
            ### load every module usable on this system
            try:
                self._modmgr.load()
            except ModuleManagerError as err:
                self._logger.error('initial module load failure: %s', err)
            
            ### wait for the sjrpc main loop to die
            while True:
                rpc_thread.join(3600)
                if not rpc_thread.is_alive():
                    self._logger.info('''sjrpc returned from it's main loop''')
                    break
                else:
                    self._logger.info('*** hourly life report ***')
        
        except:
            raise
        
        finally:
            self.shutdown()
    
    def shutdown(self):
        
        with self._mutex:
            if not self._shutdown.is_set():
                # unload every module
                # TODO currently will wait forever if a module blocks
                self._logger.warning('unloading all modules')
                self._modmgr.unload()
                # unlock main loop
                self._logger.warning('killing main loop')
                # kill sjrpc main loop
                # FIXME bad API
                self._rpc_manager.shutdown()
                self._shutdown.set()


class ModuleManagerError(Exception):
    
    pass


class ModuleManager(object):
    
    def __init__(self, logger, config):
        
        self._logger = logger
        self._config = config
        self._load_mutex = Lock()
        self._dict_mutex = RWLock()
        # maps a module name to the python module of this cc-node module :-)
        # so it usually contains all available modules
        self._mod_loaded = {}
        # maps module names to module handles (instance of initialized module)
        # so these are currently 'running' modules
        self._mod_handles = {}
        
        # Note: the only methods (load and unload) doing writes on _mod_* dicts 
        #       will lock _load_mutex, so for the sake of simplicity when we're 
        #       in a block protected by this mutex, we'll only use
        #       _dict_mutex.write as _dict_mutex.read becomes useless there
    
    def _reload_list(self):
        
        self._logger.debug(1, 'reloading modules package')
        # FIXME maybe use _reload_module and benefit from error checks
        reload(modules)
        
        if not isinstance(modules.LIST, list):
            self._logger.debug(1, 'no module list in module package')
            modules.LIST = []
    
    def _reload_module(self, name):
        
        self._logger.debug(1, 'reloading module code: %s', name)
        with self._dict_mutex.write:
            try:
                # FIXME is there a better way to handle import VS reload ?
                # FIXME do absolute import, but preserving package namespace
                #       so we can access module with 'modules.foobar'
                # FIXME code injection (don't use exec or sanitize input)
                if name not in self._mod_loaded:
                    exec('import ccnode.modules.%s' % name)
                else:
                    exec('reload(modules.%s)' % name)
                
            except ImportError:
                raise ModuleManagerError('%s: import error, the module or a'
                                    ''' dependancy was not found''' % name)
            
            except KeyboardInterrupt:
                raise
                
            except Exception:
                self._logger.backtrace()
                raise ModuleManagerError('%s: error in module code, backtrace'
                                            ' was sent to debug logging' % name)
            else:
                self._mod_loaded[name] = getattr(modules, name)

    def _mod_validate(self, name, module): 
    
        if not hasattr(module, 'Module'):
            self._logger.error('module with no definition: %s', name)
            return False
            
        elif not issubclass(module.Module, BaseModule):
            self._logger.error('module does not inherint BaseModule: %s', name)
            return False
            
        elif not (hasattr(module.Module, 'mod_dependancies') and
                                    callable(module.Module.mod_dependancies)):
            # TODO check the prototype too
            self._logger.error('module is missing the mod_dependancies'
                                                        ' method: %s', name)
            return False
            
        elif not (hasattr(module.Module, 'mod_initialize') and
                                        callable(module.Module.mod_initialize)):
            # TODO check the prototype too
            self._logger.error('module is missing the mod_initialize'
                                                        ' method: %s', name)
            return False
        
        return True

    def load(self, names=None):
        
        # read the note about locking in __init__()
        
        with self._load_mutex:
            # reload module index
            self._reload_list()
            
            # if no name is provided, we'll load all modules
            if isinstance(names, list):
                to_load = names
            elif isinstance(names, str):
                to_load = [names]
            else:
                to_load = modules.LIST
            
            for mod_name in to_load:
                ### skip name if already loaded or not in package
                if mod_name in self._mod_handles:
                    self._logger.debug(1, 'skipping module already'
                                                ' initialized: %s', mod_name)
                    # abort loading this module
                    continue
                elif mod_name not in modules.LIST:
                    self._logger.debug(1, 'not loading unindexed'
                                                    ' module: %s', mod_name)
                    # abort loading this module
                    continue
                
                ### (re)load module code
                try:
                    self._reload_module(mod_name)
                    with self._dict_mutex.write:
                        mod = self._mod_loaded[mod_name]
                        if not self._mod_validate(mod_name, mod):
                            raise ModuleManagerError('validation failed')
                except ModuleManagerError as err:
                    self._logger.error('failed to load module: %s', err)
                    # abort loading this module
                    continue
                
                ### (re)load dependancies and build deptree
                deptree = _DepTree()
                # add our current module and it's deps
                deptree.add(mod_name, mod.Module.mod_dependancies())
                # deptree tells us what to load
                load_error = False
                while not load_error:
                    # get one name to load
                    dep_name = deptree.next_missing_dep()
                    # exit if over
                    if not dep_name:
                        break
                    # (re)load it, _only if not in use_
                    try:
                        if not dep_name in self._mod_handles:
                            self._reload_module(dep_name)
                        with self._dict_mutex.write:
                            dep = self._mod_loaded[dep_name]
                            if not self._mod_validate(dep_name, dep):
                                raise ModuleManagerError('validation failed')
                    except ModuleManagerError as err:
                        self._logger.error('failed to load dependancy'
                                                            ' module: %s', err)
                        # stop the dep scan when load failed
                        load_error = True
                        continue
                    # add the deps of this new mod
                    deptree.add(dep_name, dep.Module.mod_dependancies())
                # abort all in case we couldn't load a dep
                if load_error:
                    continue
                
                ### now that all modules are loaded, we should initialize
                ### them in the right order, which is provided by _DepTree
                try:
                    for name in deptree.ordered_init_list():
                        # ONLY if not initialized yet
                        if name not in self._mod_handles:
                            self._logger.debug(1, 'initializing module: %s',
                                                                        name)
                            mod = self._mod_loaded[name]
                            # init the module
                            if not mod.Module.mod_initialize(self):
                                # module doesn't wish to load, abort
                                # (yes it can, and that's not an error)
                                self._logger.info('module refused to load: %s',
                                                                        name)
                                break
                            # add to list of initialized mods if we have an obj
                            if (hasattr(mod.Module, 'handle') and
                                    isinstance(mod.Module.handle, mod.Module)):
                                self._mod_handles[name] = mod.Module.handle
                                self._logger.info('module initialized: %s',
                                                                        name)
                            else:
                                self._logger.error('module did not provide a'
                                            ' valid instance handle: %s', name)
                                break
                # dependancy error means we couldn't init anything
                except _DepTreeError as err:
                    self._logger.error('a dependancy error prevented loading'
                                        ' of module %s: %s', mod_name, err)
                    # abort all, no module was initialized
                    continue
                # if any error occur during initialization, can't continue
                # with a missing module. we leave things as they are.
                except ModuleError as err:
                    self._logger.error('failed to initialize module %s: %s',
                                                                    name, err)
                    # abort all, hopefully the module cleaned it's ressources
                    continue
                except KeyboardInterrupt:
                    raise
                except Exception as err:
                    self._logger.error('miserable failure during module'
                        ' initialization of `%s` go check the code'
                        ' immediately, backtrace in debug log: %s', name, err)
                    self._logger.backtrace()
                    # abort all, hopefully the module cleaned it's ressources
                    continue
            
            # raise error if the job is not complete
            # FIXME should not reporte error when mod_initialize() was False
            missing = set(to_load) - set(self._mod_handles.keys())
            if len(missing):
                raise ModuleManagerError('could not load: %s' %
                                                            ', '.join(missing))

    def unload(self, names=None):
        
        # please read the note about locking in __init__()
        
        with self._load_mutex:
            if isinstance(names, list):
                to_unload = names
            elif isinstance(names, str):
                to_unload = [names]
            else:
                to_unload = self._mod_loaded.keys()
        
        # FIXME
        # check tah _loaded do not contains modules with deps to to_unload

    # FIXME should disappear, see Core.run()
    def set_core_handler(self, handler):
        
        self._core_handler = handler

    def get_handler(self):
        
        return self._core_handler
    
    def get_logger(self):
        
        return self._logger
    
    def get_configuration(self):
        
        return self._config
    
    def get(self, name):
        
        with self._dict_mutex.read:
            if name in self._mod_handles:
                return self._mod_handles[name]
            else:
                raise ModuleManagerError('access to uninitialized module: %s'
                                                                        % name)


class _DepTreeError(Exception):
    
    pass


class _DepTree(object):
    
    def __init__(self):
        
        self._mod_deps = {}
    
    def add(self, name, deps):
        
        # save it in our mod->deps mapping
        self._mod_deps[name] = deps
    
    def next_missing_dep(self):
        
        # return the first module present in a dependancies list
        # but not registered in the mod->deps mapping
        for mod, deps in self._mod_deps.iteritems():
            for dep in deps:
                if dep not in self._mod_deps:
                    return dep
        
        # or None when we do not miss anything anymore
        return None
        
    def ordered_init_list(self):
        
        #
        # How it works:
        #
        #  * first I create a directed graph of all modules linked
        #    by dependancies, despite the class name it's not always
        #    a tree
        #  * then I ensure that we have no cycle in it, or it means
        #    unresolvable deps and the process stops. actually this is
        #    done during the following step
        #  * finally the list is built starting with modules that have
        #    no child, level-by-level up to the last parent(s)
        #
        # Graph node:
        #  (mod_name, {deps name:@ref})
        #
        
        ### create the directed graph of dependancies 
        graph = []
        for mod_name, mod_deps_list in self._mod_deps.iteritems():
            # create node tuple
            mod_deps = dict([(name, None) for name in mod_deps_list])
            mod = (mod_name, mod_deps)
            # iter on nodes to make arcs
            for node in graph:
                n_name, n_deps = node
                # add node->mod arc
                if mod_name in n_deps:
                    n_deps[mod_name] = mod
                # add mod->node arc
                for dep_name, dep in mod_deps.iteritems():
                    if n_name == dep_name:
                        dep = node
            # add the new node to the graph
            graph.append(mod)
        
        ### build an ideal module initialization list that obey dependancies
        # scan for the initial level: modules with no dependancies
        level = []
        for node in graph:
            n_name, n_deps = node
            if not len(n_deps):
                level.append(node)
                
        # stop if we've lost already
        if not len(level) and len(self._mod_deps):
            raise _DepTreeError('cross-dependancies on all the modules')
            
        # now iter on next levels
        load_list = []
        while True:
            # extend the list with current level
            for n_name, n_deps in level:
                load_list.append(n_name)
            # build the next level
            next_level = []
            for node in graph:
                n_name, n_deps = node
                # only add _new_ nodes that are fulfiled with the current list
                if (n_name not in load_list
                                    and set(n_deps.keys()) <= set(load_list)):
                    next_level.append(node)
            # stop when the next level is empty
            if not len(next_level):
                # if this is a cross-deps situation, abort
                if len(load_list) < len(graph):
                    # FIXME enhance the message explaining what modules remain
                    #       to load
                    raise _DepTreeError('cross-dependancies loop')
                # we're finished
                else:
                    break
            # good to go to the next level
            else:
                level = next_level
        
        return load_list


class ModuleError(Exception):
    
    pass


class BaseModule(object):
    
    _DEPS = []

    @classmethod
    def mod_dependancies(cls):

        return cls._DEPS if isinstance(cls._DEPS, list) else []
    
    @classmethod
    def mod_initialize(cls, manager):

        if not hasattr(cls, 'loadable') or (callable(cls.loadable)
                                                            and cls.loadable()):
            cls.handle = cls(manager)
            return True
        else:
            return False
    
    @classmethod
    def mod_destroy(cls):
        
        if hasattr(self, 'destroy') and callable(self.destroy):
           self.destroy()
        del cls.handle
    
    def __init__(self, manager):
        
       if hasattr(self, 'init') and callable(self.init):
           self.init(manager) 


class HandlerError(Exception):
    
    pass


class Handler(object):
        
    def __init__(self, logger, name):
        
        self._logger = logger
        self._name = self.__class__.__name__ if name is None else name
        self._mutex = RWLock()
        self._procedures = {}
        self._children = []
    
    def __getitem__(self, name):
        
        if name.startswith('_'):
            raise AttributeError('attribute `%s` is private' % name)
        
        search = self._search_procedure(name)
        
        if search is not None:
            handler, procedure = search
            self._logger.debug(1, 'handler access for: %s->%s', handler, name)
            return procedure
        else:
            self._logger.info('procedure `%s` is not registered in subtree'
                                ' of handler `%s`' % (name, self._name))
            raise KeyError('procedure `%s` is not registered' % name)
    
    def _search_procedure(self, name):
        
        with self._mutex.read:
            # return local procedure if present, with handler name
            if name in self._procedures:
                return (self._name, self._procedures[name])
            
            # search in sub-handlers
            for handler in self._children:
                search = handler._search_procedure(name)
                if search is not None:
                    return search
            
            # not found at all
            return None
    
    def register(self, handler):
        
        with self._mutex.write:
            if handler not in self._children:
                self._children.append(handler)
            else:
                raise HandlerError('handler `%s` is already registered in `%s`'
                                        % (handler._name, self._name))
    
    def unregister(self, handler):
        
        with self._mutex.write:
            while handler in self._children:
                self._children.remove(handler)
    
    def procedure_register(self, name, procedure):
        
        with self._mutex.write:
            if name not in self._procedures:
                self._procedures[name] = procedure
            else:
                raise HandlerError('procedure `%s` is already registered'
                                            ' in `%s`' % (name, self._name))
    
    def procedure_unregister(self, name):
        
        with self._mutex.write:
            if name in self._procedures:
                self._procedures.pop(name)
