# -*- coding: utf-8 -*-
#
#  Cloud-Control Node
#  Copyright (C) 2011 SmartJog [http://www.smartjog.com]
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  Authors:
#    * 2011-06: Thibault VINCENT <thibault.vincent@smartjog.com>
#

from __future__ import absolute_import
from signal import signal, SIGHUP, SIGUSR1
from time import sleep
from .logging import Logger, DEBUG, INFO, ERROR
from .config import Configuration
from .kernel import Core
from . import __product__, __version__


class Launcher(object):
    # pylint: disable=R0903
    
    def __init__(self):
        
        # initialize core components
        self._logger = Logger()
        self._logger.info('initializing %s version %s' % (__product__,
                                                            str(__version__)))
        self._config = Configuration(self._logger)
        self._core = Core(self._logger, self._config)
    
    def _on_sighup(self, signum, frame):
        # pylint: disable=W0613
        
        self._logger.warning('SIGHUP received, shutting down core')
        self._core.shutdown()
    
    def _on_sigusr1(self, signum, frame):
        # pylint: disable=W0613
        
        self._logger.info('SIGUSR1 received, reloading configuration file and'
                                                            ' logging settings')
        self._reload_conf()
    
    def _reload_conf(self):
        
        # (re)load configuration file
        self._config.parse_file()
        
        # (re)apply logging configuration
        # verbosity level
        verbosity = self._config.get('verbosity')
        if verbosity >= 1:
            self._logger.set_level(DEBUG)
            self._logger.set_debug_level(verbosity)
        else:
            self._logger.set_level(INFO)
        # stderr/syslog output selection
        std_output = self._config.get('stdout')
        self._logger.set_stderr_enabled(std_output)
        self._logger.set_syslog_enabled(not std_output)
    
    def main(self):
        
        # read command line arguments
        self._config.parse_arguments()
        
        # register hangup signal to restart core
        signal(SIGHUP, self._on_sighup)
        
        # register USR1 signal to reload configuration file
        # and apply logging settings
        signal(SIGUSR1, self._on_sigusr1)
        
        # run the core untill permanent shutdown is requested
        while True:
            try:
                # (re)load configuration file and apply logging settings
                self._reload_conf()
                
                # start core main loop
                self._core.run()
                sleep(0.5)
            
            except KeyboardInterrupt:
                self._logger.warning('SIGINT received, shutting down core'
                                                                ' post-mortem')
                self._core.shutdown()
                break
            
            except Exception:
                self._logger.error('unexpected error raised by core')
                self._logger.backtrace(priority=ERROR)
                sleep(5)
