# -*- coding: utf-8 -*-
#
#  Cloud-Control Node
#  Copyright (C) 2011 SmartJog [http://www.smartjog.com]
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  Authors:
#    * 2011-06: Thibault VINCENT <thibault.vincent@smartjog.com>
#

from __future__ import absolute_import
import syslog
import socket
from sys import stderr
from threading import Lock
from traceback import format_exc
from . import __canonical__, __develmode__

DEBUG = syslog.LOG_DEBUG
INFO = syslog.LOG_INFO
WARNING = syslog.LOG_WARNING
ERROR = syslog.LOG_ERR
CRITICAL = syslog.LOG_CRIT

ALL_PRIO = (DEBUG, INFO, WARNING, ERROR, CRITICAL)
PRIO_TAGS = {
    DEBUG   : 'D',
    INFO    : 'I',
    WARNING : 'W',
    ERROR   : 'E',
    CRITICAL: 'C',
}


class Logger(object):
    
    def __init__(self):
        
        self._max_prio = INFO
        self._max_debug = 1
        self._mutex = Lock()
        self._mcast_sock = None
        self._mcast_dst = None
        self._syslog_enabled = not __develmode__
        self._stderr_enabled = __develmode__
        syslog.openlog(__canonical__,
                syslog.LOG_PID ^ syslog.LOG_NDELAY ^ syslog.LOG_CONS,
                syslog.LOG_DAEMON)
    
    def set_level(self, priority):
        
        if priority not in ALL_PRIO:
            raise ValueError('not an allowed logging priority')
        
        self._max_prio = priority
    
    def set_debug_level(self, level):
        
        if level <= 0:
            raise ValueError('debug level is a positive integer')
        
        self._max_debug = int(level)
    
    def set_syslog_enabled(self, state):
        
        self._syslog_enabled = bool(state)
    
    def set_stderr_enabled(self, state):
        
        self._stderr_enabled = bool(state)
    
    def mcast_enable(self, addr, port):
        
        with self._mutex:
            self._mcast_dst = (addr, port)
            self._mcast_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM,
                                                            socket.IPPROTO_UDP)
            self._mcast_sock.setsockopt(socket.IPPROTO_IP,
                                                    socket.IP_MULTICAST_TTL, 2)
    
    def mcast_disable(self):
        
        with self._mutex:
            self._mcast_sock = None
            self._mcast_dst = None
    
    def _log(self, priority, level, message, *args):
        
        if priority not in ALL_PRIO:
            priority = CRITICAL
        
        printable = (priority <= self._max_prio and (level is None
                        or (priority == DEBUG and level <= self._max_debug)))
        
        if printable or self._mcast_sock is not None:
            try:
                if len(args):
                    message = message % args
            
            except TypeError:
                message = '[PARTIAL] %s' % message
            
            finally:
                if printable:
                    if self._syslog_enabled:
                        syslog.syslog(priority, message)
                        
                    if self._stderr_enabled:
                        if isinstance(level, int):
                            my_lvl = str(level)
                        else:
                            my_lvl = ' '
                        stderr.write('%s[%s%s]: %s\n' % (__canonical__,
                            PRIO_TAGS[priority], my_lvl, message))
                
                with self._mutex:
                    if self._mcast_sock is not None:
                        self._mcast_sock.sendto('%i %s' %
                                            (level, message), self._mcast_dst)
    
    def debug(self, level, message, *args):
        
        self._log(DEBUG, level, message, *args)
    
    def info(self, message, *args):
        
        self._log(INFO, None, message, *args)
    
    def warning(self, message, *args):
        
        self._log(WARNING, None, message, *args)

    def error(self, message, *args):
        
        self._log(ERROR, None, message, *args)
    
    def critical(self, message, *args):
        
        self._log(CRITICAL, None, message, *args)

    def backtrace(self, priority=DEBUG, level=1):
        
        if priority != DEBUG:
            level = None
        
        for line in format_exc().splitlines():
            self._log(priority, level, 'BT: %s', line)
